"""
StubCsi can be used for testing without a backing Pharia Kernel instance.
"""

from collections.abc import Generator
from typing import Sequence

from pharia_skill import (
    ChatParams,
    ChatRequest,
    ChatResponse,
    Chunk,
    ChunkRequest,
    Completion,
    CompletionParams,
    CompletionRequest,
    Csi,
    Cursor,
    Document,
    DocumentPath,
    ExplanationRequest,
    FinishReason,
    JsonSerializable,
    Language,
    Message,
    SearchRequest,
    SearchResult,
    SelectLanguageRequest,
    Text,
    TextScore,
    TokenUsage,
)
from pharia_skill.csi.inference import (
    ChatEvent,
    ChatStreamResponse,
    CompletionAppend,
    CompletionEvent,
    CompletionStreamResponse,
    MessageAppend,
    MessageBegin,
)


class StubCsi(Csi):
    """
    The `StubCsi` can be used to mock out the CSI for testing purposes.

    You can use this class directly, or inherit from it and load it up with your own expectations.
    Suppose you want to test a Skill that uses the `chat` method, and want to mock out the response from the LLM
    to run your tests faster:

    Example::

        from pharia_skill import Csi, skill

        @skill
        def run(csi: Csi, input: Input) -> Output:
            system = Message.system("You are a poet who strictly speaks in haikus.")
            user = Message.user(input.topic)
            params = ChatParams(max_tokens=64)
            response = csi.chat("llama-3.1-8b-instruct", [system, user], params)
            return Output(haiku=response.message.content.strip())

        class CustomMockCsi(StubCsi):
            def chat(self, model: str, messages: list[Message], params: ChatParams) -> ChatResponse:
                message = Message.assistant("Whispers in the dark\nEchoes of a fleeting dream\nMeaning lost in space")
                return ChatResponse(message=message, finish_reason=FinishReason.STOP)

        def test_run():
            csi = CustomMockCsi()
            result = run(csi, Input(topic="The meaning of life"))
            assert result.haiku == "Whispers in the dark\nEchoes of a fleeting dream\nMeaning lost in space"
    """

    def completion_stream(
        self, model: str, prompt: str, params: CompletionParams
    ) -> CompletionStreamResponse:
        class StubCompletionStreamResponse(CompletionStreamResponse):
            def __init__(self, stream: Generator[CompletionEvent, None, None]):
                self._stream = stream

            def next(self) -> CompletionEvent | None:
                return next(self._stream, None)

        def generator() -> Generator[CompletionEvent, None, None]:
            for char in prompt:
                yield CompletionAppend(char, [])
            yield FinishReason.STOP
            yield TokenUsage(len(prompt), len(prompt))

        return StubCompletionStreamResponse(generator())

    def chat_stream(
        self, model: str, messages: list[Message], params: ChatParams
    ) -> ChatStreamResponse:
        class StubChatStreamResponse(ChatStreamResponse):
            def __init__(self, stream: Generator[ChatEvent, None, None]):
                self._stream = stream
                super().__init__()

            def next(self) -> ChatEvent | None:
                return next(self._stream, None)

        def generator() -> Generator[ChatEvent, None, None]:
            total_usage = 0
            yield MessageBegin("assistant")
            for message in messages:
                content = message.content
                total_usage += len(content)
                yield MessageAppend(content, [])
            yield FinishReason.STOP
            yield TokenUsage(total_usage, total_usage)

        return StubChatStreamResponse(generator())

    def complete_concurrent(
        self, requests: Sequence[CompletionRequest]
    ) -> list[Completion]:
        return [
            Completion(
                text=request.prompt,
                finish_reason=FinishReason.STOP,
                logprobs=[],
                usage=TokenUsage(
                    prompt=len(request.prompt), completion=len(request.prompt)
                ),
            )
            for request in requests
        ]

    def chunk_concurrent(self, requests: Sequence[ChunkRequest]) -> list[list[Chunk]]:
        return [[Chunk(text=request.text, character_offset=0)] for request in requests]

    def chat_concurrent(self, requests: Sequence[ChatRequest]) -> list[ChatResponse]:
        return [
            ChatResponse(
                message=Message.assistant(request.messages[-1].content),
                finish_reason=FinishReason.STOP,
                logprobs=[],
                usage=TokenUsage(
                    prompt=len(request.messages[-1].content),
                    completion=len(request.messages[-1].content),
                ),
            )
            for request in requests
        ]

    def explain_concurrent(
        self, requests: Sequence[ExplanationRequest]
    ) -> list[list[TextScore]]:
        return [[] for _ in requests]

    def select_language_concurrent(
        self, requests: Sequence[SelectLanguageRequest]
    ) -> list[Language | None]:
        return [
            request.languages[0] if request.languages else None for request in requests
        ]

    def search_concurrent(
        self, requests: Sequence[SearchRequest]
    ) -> list[list[SearchResult]]:
        return [
            [
                SearchResult(
                    document_path=DocumentPath(
                        namespace="dummy-namespace",
                        collection="dummy-collection",
                        name="dummy-name",
                    ),
                    content="dummy-content",
                    score=1.0,
                    start=Cursor(item=0, position=0),
                    end=Cursor(item=0, position=0),
                )
            ]
            for _ in requests
        ]

    def documents(self, document_paths: Sequence[DocumentPath]) -> list[Document]:
        return [
            Document(
                path=document_path,
                contents=[Text("dummy-content")],
                metadata=None,
            )
            for document_path in document_paths
        ]

    def documents_metadata(
        self, document_paths: Sequence[DocumentPath]
    ) -> list[JsonSerializable]:
        return [{} for _ in document_paths]



"""
Translation between SDK types and the serialized format expected by the Pharia Kernel csi-shell endpoint.

While we could use the SDK types that we expose as part of the SDK for serialization/deserialization,
uncoupling these interfaces brings two advantages:

1. We can rename members at any time in the SDK (just a version bump) without requiring a new wit world / new version of the csi-shell.
2. We can use Pydantic models for serialization/deserialization without exposing these to the SDK users. We prefer dataclasses as they do not require keyword arguments for setup.
"""

import json
from collections.abc import Generator
from typing import Any, Sequence

from opentelemetry import trace
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.trace import StatusCode

from pharia_skill import (
    ChatParams,
    ChatRequest,
    ChatResponse,
    Chunk,
    ChunkRequest,
    Completion,
    CompletionParams,
    CompletionRequest,
    Csi,
    Document,
    DocumentPath,
    ExplanationRequest,
    JsonSerializable,
    Language,
    Message,
    SearchRequest,
    SearchResult,
    SelectLanguageRequest,
    TextScore,
)
from pharia_skill.csi.inference import ChatStreamResponse, CompletionStreamResponse
from pharia_skill.studio import (
    StudioClient,
    StudioExporter,
    StudioSpanProcessor,
)

from .chunking import ChunkDeserializer, ChunkRequestSerializer
from .client import Client, CsiClient, Event
from .document_index import (
    DocumentDeserializer,
    DocumentMetadataDeserializer,
    DocumentMetadataSerializer,
    DocumentSerializer,
    SearchRequestSerializer,
    SearchResultDeserializer,
)
from .inference import (
    ChatListDeserializer,
    ChatRequestListSerializer,
    ChatRequestSerializer,
    CompletionListDeserializer,
    CompletionRequestListSerializer,
    CompletionRequestSerializer,
    DevChatStreamResponse,
    DevCompletionStreamResponse,
    ExplanationListDeserializer,
    ExplanationRequestListSerializer,
)
from .language import (
    SelectLanguageDeserializer,
    SelectLanguageRequestSerializer,
)


class DevCsi(Csi):
    """
    The `DevCsi` can be used for testing Skill code locally against a running Pharia Kernel.

    This implementation of Cognitive System Interface (CSI) is backed by a running instance of Pharia Kernel via HTTP.
    This enables skill developers to run and test Skills against the same services that are used by the Pharia Kernel.

    Examples::

        # import your skill
        from haiku import run

        # create a `CSI` instance, optionally with trace export to Studio
        csi = DevCsi().with_studio("my-project")

        # Run your skill
        input = Input(topic="The meaning of life")
        result = run(csi, input)

        assert "42" in result.haiku

    The following environment variables are required:

    * `PHARIA_AI_TOKEN` (Pharia AI token)
    * `PHARIA_KERNEL_ADDRESS` (Pharia Kernel endpoint; example: "https://pharia-kernel.product.pharia.com")

    If you want to export traces to Pharia Studio, also set:

    * `PHARIA_STUDIO_ADDRESS` (Pharia Studio endpoint; example: "https://pharia-studio.aleph-alpha.stackit.run")
    """

    def __init__(self) -> None:
        self.client: CsiClient = Client()

    def completion_stream(
        self, model: str, prompt: str, params: CompletionParams
    ) -> CompletionStreamResponse:
        body = CompletionRequestSerializer(
            model=model, prompt=prompt, params=params
        ).model_dump()
        events = self.stream("completion_stream", body)
        return DevCompletionStreamResponse(events)

    def chat_stream(
        self, model: str, messages: list[Message], params: ChatParams
    ) -> ChatStreamResponse:
        body = ChatRequestSerializer(
            model=model, messages=messages, params=params
        ).model_dump()
        events = self.stream("chat_stream", body)
        return DevChatStreamResponse(events)

    def complete_concurrent(
        self, requests: Sequence[CompletionRequest]
    ) -> list[Completion]:
        body = CompletionRequestListSerializer(root=requests).model_dump()
        output = self.run("complete", body)
        return CompletionListDeserializer(root=output).root

    def chat_concurrent(self, requests: Sequence[ChatRequest]) -> list[ChatResponse]:
        body = ChatRequestListSerializer(root=requests).model_dump()
        output = self.run("chat", body)
        return ChatListDeserializer(root=output).root

    def explain_concurrent(
        self, requests: Sequence[ExplanationRequest]
    ) -> list[list[TextScore]]:
        body = ExplanationRequestListSerializer(root=requests).model_dump()
        output = self.run("explain", body)
        return ExplanationListDeserializer(root=output).root

    def chunk_concurrent(self, requests: Sequence[ChunkRequest]) -> list[list[Chunk]]:
        body = ChunkRequestSerializer(root=requests).model_dump()
        output = self.run("chunk_with_offsets", body)
        return ChunkDeserializer(root=output).root

    def select_language_concurrent(
        self, requests: Sequence[SelectLanguageRequest]
    ) -> list[Language | None]:
        body = SelectLanguageRequestSerializer(root=requests).model_dump()
        output = self.run("select_language", body)
        return SelectLanguageDeserializer(root=output).root

    def search_concurrent(
        self, requests: Sequence[SearchRequest]
    ) -> list[list[SearchResult]]:
        body = SearchRequestSerializer(root=requests).model_dump()
        output = self.run("search", body)
        return SearchResultDeserializer(root=output).root

    def documents_metadata(
        self, document_paths: Sequence[DocumentPath]
    ) -> list[JsonSerializable | None]:
        body = DocumentMetadataSerializer(root=document_paths).model_dump()
        output = self.run("document_metadata", body)
        return DocumentMetadataDeserializer(root=output).root

    def documents(self, document_paths: Sequence[DocumentPath]) -> list[Document]:
        body = DocumentSerializer(root=document_paths).model_dump()
        output = self.run("documents", body)
        return DocumentDeserializer(root=output).root

    @classmethod
    def with_studio(cls, project: str) -> "DevCsi":
        """Create a `DevCsi` that exports traces to Pharia Studio.

        This function creates a `StudioExporter` and registers it with the tracer provider.
        The exporter uploads spans once the root span ends.

        Args:
            project: The name of the studio project to export traces to. Will be created if it does not exist.
        """
        csi = cls()
        client = StudioClient.with_project(project)
        exporter = StudioExporter(client)
        csi.set_span_exporter(exporter)
        return csi

    @classmethod
    def set_span_exporter(cls, exporter: StudioExporter) -> None:
        """Set a span exporter for Studio if it has not been set yet.

        This method overwrites any existing exporters, thereby ensuring that there
        are never two exporters to Studio attached at the same time.
        """
        provider = cls.provider()
        for processor in provider._active_span_processor._span_processors:
            if isinstance(processor, StudioSpanProcessor):
                processor.span_exporter = exporter
                return

        span_processor = StudioSpanProcessor(exporter)
        provider.add_span_processor(span_processor)

    @classmethod
    def existing_exporter(cls) -> StudioExporter | None:
        """Return the first studio exporter attached to the provider, if any."""
        provider = cls.provider()
        for processor in provider._active_span_processor._span_processors:
            if isinstance(processor, StudioSpanProcessor):
                if isinstance(processor.span_exporter, StudioExporter):
                    return processor.span_exporter
        return None

    @staticmethod
    def provider() -> TracerProvider:
        """Tracer provider for the current thread.

        Check if the tracer provider is already set and if not, set it.
        """
        if not isinstance(trace.get_tracer_provider(), TracerProvider):
            trace_provider = TracerProvider()
            trace.set_tracer_provider(trace_provider)

        return trace.get_tracer_provider()  # type: ignore

    def run(self, function: str, data: dict[str, Any]) -> Any:
        with trace.get_tracer(__name__).start_as_current_span(function) as span:
            span.set_attribute("input", json.dumps(data))
            try:
                output = self.client.run(function, data)
            except Exception as e:
                span.set_status(StatusCode.ERROR, str(e))
                raise e
            span.set_attribute("output", json.dumps(output))

        return output

    def stream(
        self, function: str, data: dict[str, Any]
    ) -> Generator[Event, None, None]:
        with trace.get_tracer(__name__).start_as_current_span(function) as span:
            span.set_attribute("input", json.dumps(data))
            # We don't know if the entire stream will be consumed, so set status to OK here
            try:
                events = self.client.stream(function, data)
                for event in events:
                    # OpenTelemetry only allows for one level of nesting in attributes, hence we flatten
                    # nested dicts to json strings.
                    # Studio itself would support arbitray nesting (see `StudioSpan.events`), but as we use
                    # OpenTelemetry for tracing and only convert to Studio spans when exporting, we need to
                    # respeect that limitation here.
                    # Another option would be to load these json strings into Python dicts in `StudioSpan.from_otel`.
                    attributes = {
                        k: json.dumps(v) if isinstance(v, dict) else v
                        for k, v in event.data.items()
                    }
                    span.add_event(event.event, attributes=attributes)
                    if event.event == "error":
                        raise ValueError(event.data["message"])
                    yield event
            except Exception as e:
                span.set_status(StatusCode.ERROR, str(e))
                raise e



"""
The protocol we offer to Skill developers.

This interface is type-checked, in that we validate all data that we pass along to the
bindings. For this validation, we rely on `pydantic.dataclasses` to validate the types.

Python uses duck typing and this implies that someone using a library is responsible to
ensure the types that are passed are correct. This assumption holds true for all types that
we use in the SDK, and do not pass on to the bindings. However, it is the responsibility of
the SDK to ensure that the error messages users receive if they pass incorrect types are
are actionable. With `componentize-py`, the backtrace which is offered does not guide the
user towards the correct solution. Therefore, it is our responsibility to make sure that
we pass the correct types to the bindings.

This validation responsibility only covers types we pass to the bindings, it does not extend
to assert the type of something we use in our Python code. E.g. if we receive a DocumentPath
in our `Csi` interface, and only use this to instantiate a `bindings.imports.DocumentPath`,
the Python interpreter makes sure the caller gets a good error message if they pass in e.g.
`None` and we access the `name` attribute in our SDK.
"""

from typing import Protocol, Sequence

from .chunking import Chunk, ChunkParams, ChunkRequest
from .document_index import (
    Document,
    DocumentPath,
    IndexPath,
    JsonSerializable,
    SearchFilter,
    SearchRequest,
    SearchResult,
)
from .inference import (
    ChatParams,
    ChatRequest,
    ChatResponse,
    ChatStreamResponse,
    Completion,
    CompletionParams,
    CompletionRequest,
    CompletionStreamResponse,
    ExplanationRequest,
    Granularity,
    Message,
    TextScore,
)
from .language import Language, SelectLanguageRequest


class Csi(Protocol):
    """The Cognitive System Interface (CSI) is a protocol that allows skills to interact with the Pharia Kernel.

    Most functionality in the CSI is offered in two forms: As a single request, and as multiple
    concurrent requests. For all concurrent requests, it is guaranteed that the responses are
    returned in the same order as the requests. Therefore, our interface requires the user to provide
    Sequences, as we want the input to be ordered.
    """

    def completion_stream(
        self, model: str, prompt: str, params: CompletionParams
    ) -> CompletionStreamResponse:
        """Streams completions given a prompt.

        Parameters:
            model (str, required): Name of model to use.
            prompt (str, required): The text to be completed.
            params (CompletionParams, required): Parameters for the requested completion.
        """
        ...

    def chat_stream(
        self, model: str, messages: list[Message], params: ChatParams
    ) -> ChatStreamResponse:
        """Streams chat with a model, where the first item is the role.

        Parameters:
            model (str, required):
                Name of model to use.

            messages (list[Message], required):
                List of messages, alternating between messages from user and assistant.

            params (ChatParams, required):
                Parameters used for the chat.
        """
        ...

    def complete(self, model: str, prompt: str, params: CompletionParams) -> Completion:
        """Generates completions given a prompt.

        Parameters:
            model (str, required): Name of model to use.
            prompt (str, required): The text to be completed.
            params (CompletionParams, required): Parameters for the requested completion.

        Examples::

            prompt = f\"\"\"<|begin_of_text|><|start_header_id|>system<|end_header_id|>

            You are a poet who strictly speaks in haikus.<|eot_id|><|start_header_id|>user<|end_header_id|>

            {input.root}<|eot_id|><|start_header_id|>assistant<|end_header_id|>\"\"\"
            params = CompletionParams(max_tokens=64)
            completion = csi.complete("llama-3.1-8b-instruct", prompt, params)
        """
        request = CompletionRequest(model, prompt, params)
        return self.complete_concurrent([request])[0]

    def complete_concurrent(
        self, requests: list[CompletionRequest]
    ) -> list[Completion]:
        """Generate completions concurrently.

        Parameters:
            requests (list[CompletionRequest], required): List of completion requests.
        """
        ...

    def chunk(self, text: str, params: ChunkParams) -> list[Chunk]:
        """Chunks a text into chunks according to params.

        Parameters:
            text (str, required): Text to be chunked.
            params (ChunkParams, required): Parameter used for chunking, model and maximal number of tokens.

        Examples::

            text = "A very very very long text that can be chunked."
            params = ChunkParams("llama-3.1-8b-instruct", max_tokens=5)
            result = csi.chunk(text, params)
            assert len(result) == 3
        """
        request = ChunkRequest(text, params)
        return self.chunk_concurrent([request])[0]

    def chunk_concurrent(self, requests: Sequence[ChunkRequest]) -> list[list[Chunk]]:
        """Chunk a text into chunks concurrently.

        Parameters:
            requests (list[ChunkRequest], required): List of chunk requests.
        """
        ...

    def chat(
        self, model: str, messages: list[Message], params: ChatParams
    ) -> ChatResponse:
        """Chat with a model.

        Parameters:
            model (str, required):
                Name of model to use.

            messages (list[Message], required):
                List of messages, alternating between messages from user and assistant.

            params (ChatParams, required):
                Parameters used for the chat.

        Examples::

            input = "oat milk"
            msg = Message.user(f\"\"\"You are a poet who strictly speaks in haikus.

            {input}\"\"\")
            model = "llama-3.1-8b-instruct"
            chat_response = csi.chat(model, [msg], ChatParams(max_tokens=64))
        """
        request = ChatRequest(model, messages, params)
        return self.chat_concurrent([request])[0]

    def chat_concurrent(self, requests: Sequence[ChatRequest]) -> list[ChatResponse]:
        """Chat with a model concurrently.

        Parameters:
            requests (list[ChatRequest], required): List of chat requests.
        """
        ...

    def explain(
        self,
        prompt: str,
        target: str,
        model: str,
        granularity: Granularity = Granularity.AUTO,
    ) -> list[TextScore]:
        request = ExplanationRequest(prompt, target, model, granularity)
        return self.explain_concurrent([request])[0]

    def explain_concurrent(
        self, requests: Sequence[ExplanationRequest]
    ) -> list[list[TextScore]]: ...

    def select_language(self, text: str, languages: list[Language]) -> Language | None:
        """Select the detected language for the provided input based on the list of possible languages.

        If no language matches, None is returned.

        Parameters:
            text (str, required): Text input.
            languages (list[Language], required): All languages that should be considered during detection.

        Examples::

            text = "Ich spreche Deutsch nur ein bisschen."
            languages = [Language.English, Language.German]
            result = csi.select_language(text, languages)
        """
        request = SelectLanguageRequest(text, languages)
        return self.select_language_concurrent([request])[0]

    def select_language_concurrent(
        self, requests: Sequence[SelectLanguageRequest]
    ) -> list[Language | None]:
        """Detect the language for multiple texts concurrently.

        Parameters:
            requests (list[SelectLanguageRequest], required): List of select language requests.
        """
        ...

    def search(
        self,
        index_path: IndexPath,
        query: str,
        max_results: int = 1,
        min_score: float | None = None,
        filters: list[SearchFilter] | None = None,
    ) -> list[SearchResult]:
        """Search an existing Index in the Document Index.

        Parameters:
            index_path (IndexPath, required): Index path in the Document Index to access.
            query (str, required): Text to be search for.
            max_results (int, required): Maximal number of results.
            min_score (float, optional, Default NoneNone): Minimal score for result to be included.

        Examples::

            index_path = IndexPath("f13", "wikipedia-de", "luminous-base-asymmetric-64")
            query = "What is the population of Heidelberg?"
            result = csi.search(index_path, query)
            r0 = result[0]
            "Heidelberg" in r0.content, "Heidelberg" in r0.document_path.name # True, True
        """
        request = SearchRequest(
            index_path, query, max_results, min_score, filters or []
        )
        return self.search_concurrent([request])[0]

    def search_concurrent(
        self, requests: Sequence[SearchRequest]
    ) -> list[list[SearchResult]]: ...

    """Execute multiple search requests against the Document Index."""

    def document(self, document_path: DocumentPath) -> Document:
        """Fetch a document from the Document Index.

        Parameters:
            document_path (DocumentPath, required): The document path to get the document from.

        Examples::

            document_path = DocumentPath("f13", "wikipedia-de", "Heidelberg")
            document = csi.document(document_path)
            assert document.path == document_path
        """
        return self.documents([document_path])[0]

    def documents(self, document_paths: Sequence[DocumentPath]) -> list[Document]:
        """Fetch multiple documents from the Document Index.

        The documents are guaranteed to be returned in the same order as the document paths.

        Parameters:
            document_paths (list[DocumentPath], required): The document paths to get the documents from.
        """
        ...

    def document_metadata(self, document_path: DocumentPath) -> JsonSerializable:
        """Return the metadata of a document in the Document Index.

        Parameters:
            document_path (DocumentPath, required): The document path to get metadata from.
        """
        return self.documents_metadata([document_path])[0]

    def documents_metadata(
        self, document_paths: Sequence[DocumentPath]
    ) -> list[JsonSerializable]:
        """Return the metadata of multiple documents in the Document Index.

        The metadata is guaranteed to be returned in the same order as the document paths.

        Parameters:
            document_paths (list[DocumentPath], required): The document paths to get metadata from.
        """
        ...



from ..bindings.imports import chunking as wit
from ..csi.chunking import Chunk, ChunkParams, ChunkRequest


def chunk_params_to_wit(chunk_params: ChunkParams) -> wit.ChunkParams:
    return wit.ChunkParams(
        model=chunk_params.model,
        max_tokens=chunk_params.max_tokens,
        overlap=chunk_params.overlap,
    )


def chunk_request_to_wit(
    chunk_request: ChunkRequest,
) -> wit.ChunkWithOffsetRequest:
    return wit.ChunkWithOffsetRequest(
        text=chunk_request.text,
        params=chunk_params_to_wit(chunk_request.params),
        character_offsets=True,
    )


def chunk_from_wit(chunk: wit.ChunkWithOffset) -> Chunk:
    # The character offset is always expected becaused the flag `character_offsets` is enabled
    assert chunk.character_offset is not None
    return Chunk(text=chunk.text, character_offset=chunk.character_offset)



from typing import TypeVar, Generic, Union, Optional, Protocol, Tuple, List, Any, Self
from types import TracebackType
from enum import Flag, Enum, auto
from dataclasses import dataclass
from abc import abstractmethod
import weakref

from ..types import Result, Ok, Err, Some
from ..exports import skill_handler
from ..imports import streaming_output

class SkillHandler(Protocol):

    @abstractmethod
    def run(self, input: bytes) -> bytes:
        """
        Raises: `bindings.types.Err(bindings.imports.skill_handler.Error)`
        """
        raise NotImplementedError

    @abstractmethod
    def metadata(self) -> skill_handler.SkillMetadata:
        raise NotImplementedError


class MessageStream(Protocol):

    @abstractmethod
    def run(self, input: bytes, output: streaming_output.StreamOutput) -> None:
        """
        Run the skill. Output is streamed out of the stream-output resource.
        The skill is allowed to also terminate early with an error.
        
        Raises: `bindings.types.Err(bindings.imports.message_stream.Error)`
        """
        raise NotImplementedError





from typing import Sequence

from pydantic import RootModel

from pharia_skill.csi import Language, SelectLanguageRequest

SelectLanguageRequestSerializer = RootModel[Sequence[SelectLanguageRequest]]


SelectLanguageDeserializer = RootModel[list[Language | None]]



from typing import TypeVar, Generic, Union, Optional, Protocol, Tuple, List, Any, Self
from types import TracebackType
from enum import Flag, Enum, auto
from dataclasses import dataclass
from abc import abstractmethod
import weakref

from ..types import Result, Ok, Err, Some



@dataclass
class Error_Internal:
    value: str


@dataclass
class Error_InvalidInput:
    value: str


Error = Union[Error_Internal, Error_InvalidInput]
"""
The set of errors which may be raised by functions in this interface
"""





import json
import re
from typing import Any, Literal

from pydantic import BaseModel


class Tool(BaseModel):
    """Provide a tool definition as a Pydantic model.

    The name of the class will be used as function name. The description of the
    function is taken from the docstring of the class. The parameters are
    specified as attributes of the model. Type hints and default arguments can
    be used to specify the schema, and a description of a parameter can be added
    with the `Field` class.

    Example::

        from pydantic import BaseMode, Field

        class GetImageInformation(BaseModel):
            "Retrieve information about a specific image."

            registry: str
            repository: str = Field(
                description="The full identifier of the image in the registry",
            )
            tag: str = "latest"
    """

    @classmethod
    def json_schema(cls) -> dict[str, Any]:
        """The (slightly incompliant) json schema of a tool.

        This schema is used in two ways:

        1. Passed to the model to define the tool
        2. For serialization to json as part of a chat request

        For all specified tools, this schema is passed to the model.
        LLama expects a json object with `type` "function" as the root elements
        and a `function` object with the keys `name`, `description`, and `parameters`.

        Only for the parameters, we can make use of the json schema representation of a
        pydantic models. Note that the output schema is invalid json schema, as there is
        no `function` type in the json schema specification:

        https://json-schema.org/draft/2020-12/json-schema-validation#section-6.1.1
        """
        schema = cls.model_json_schema()
        description = schema.get("description")
        if description is not None:
            del schema["description"]
        data = {
            "type": "function",
            "function": {
                "name": cls.name(),
                "description": description,
                "parameters": schema,
            },
        }
        cls._recursive_purge_title(data)
        return data

    @classmethod
    def name(cls) -> str:
        return cls._to_snake_case(cls.__name__)

    @classmethod
    def _to_snake_case(cls, name: str) -> str:
        return re.sub(r"(?<!^)(?=[A-Z])", "_", name).lower()

    @classmethod
    def _recursive_purge_title(cls, data: dict[str, Any]) -> None:
        """Remove the title field from a dictionary recursively.

        The title is automatically created based on the name of the pydantic model,
        but it is not shown in examples of the llama model card, hence we skip it.
        See https://github.com/pydantic/pydantic/discussions/8504 for more detail.
        """
        if isinstance(data, dict):
            for key in list(data.keys()):
                if key == "title" and "type" in data.keys():
                    del data[key]
                else:
                    cls._recursive_purge_title(data[key])

    def render(self) -> str:
        """Convert a tool call to prompt format again.

        When a tool call has been loaded from a model response, it is part of the
        conversation and needs to be converted back to a prompt when providing the
        full conversation history to the model for the next turn.
        """
        return json.dumps(
            {
                "name": self.name(),
                "parameters": self.model_dump(exclude_unset=True),
            }
        )


class Function(BaseModel):
    name: str
    parameters: dict[str, Any]
    description: str | None = None


class JsonSchema(BaseModel):
    """Provide a tool definition as a json schema.

    While `Tool` is a more user-friendly way to define a tool in
    code, in some cases it might put too many constraints on the user. E.g., it can
    not be serialized from a json http request. Therefore, function definitions can
    also be provided in the serialized, json schema format.
    """

    type: Literal["function"] = "function"
    function: Function

    def name(self) -> str:
        return self.function.name


ToolDefinition = type[Tool] | JsonSchema
"""A tool can either be defined as a Pydantic model or directly as a json schema."""


class CodeInterpreter(Tool):
    src: str

    def render(self) -> str:
        return self.src

    def run(self) -> Any:
        global_vars: dict[str, Any] = {}
        exec(self.src, global_vars)
        return global_vars.get("result")

    @classmethod
    def json_schema(cls) -> dict[str, Any]:
        """Json representation of the code interpreter tool.

        This is not passed to the model, but only used for serialization to json
        as part of a chat request.
        """
        return {"type": "code_interpreter"}


class WolframAlpha(Tool):
    query: str

    def render(self) -> str:
        return f'wolfram_alpha.call(query="{self.query}")'

    @staticmethod
    def try_from_text(text: str) -> "WolframAlpha | None":
        if not text.startswith("wolfram_alpha.call"):
            return None
        try:
            query = text.split('wolfram_alpha.call(query="')[1].split('")')[0].strip()
            return WolframAlpha(query=query)
        except IndexError:
            return None

    @classmethod
    def json_schema(cls) -> dict[str, Any]:
        """Json representation of the wolfram alpha tool.

        This is not passed to the model, but only used for serialization to json
        as part of a chat request.
        """
        return {"type": "wolfram_alpha"}


class BraveSearch(Tool):
    query: str

    def render(self) -> str:
        return f'brave_search.call(query="{self.query}")'

    @staticmethod
    def try_from_text(text: str) -> "BraveSearch | None":
        if not text.startswith("brave_search.call"):
            return None
        try:
            query = text.split('brave_search.call(query="')[1].split('")')[0].strip()
            return BraveSearch(query=query)
        except IndexError:
            return None

    @classmethod
    def json_schema(cls) -> dict[str, Any]:
        """Json representation of the brave search tool.

        This is not passed to the model, but only used for serialization to json
        as part of a chat request.
        """
        return {"type": "brave_search"}


BuiltInTools: tuple[type[Tool], ...] = (CodeInterpreter, WolframAlpha, BraveSearch)



import json
from typing import Sequence

from pharia_skill.csi.inference import ChatStreamResponse, CompletionStreamResponse

from ..bindings.imports import chunking as wit_chunking
from ..bindings.imports import document_index as wit_document_index
from ..bindings.imports import inference as wit_inference
from ..bindings.imports import language as wit_language
from ..csi import (
    ChatParams,
    ChatRequest,
    ChatResponse,
    Chunk,
    ChunkRequest,
    Completion,
    CompletionParams,
    CompletionRequest,
    Csi,
    Document,
    DocumentPath,
    ExplanationRequest,
    JsonSerializable,
    Language,
    Message,
    SearchRequest,
    SearchResult,
    SelectLanguageRequest,
    TextScore,
)
from .chunking import chunk_from_wit, chunk_request_to_wit
from .document_index import (
    document_from_wit,
    document_path_to_wit,
    search_request_to_wit,
    search_result_from_wit,
)
from .inference import (
    WitChatStreamResponse,
    WitCompletionStreamResponse,
    chat_request_to_wit,
    chat_response_from_wit,
    completion_from_wit,
    completion_request_to_wit,
    explanation_request_to_wit,
    text_score_from_wit,
)
from .language import language_from_wit, language_request_to_wit


class WitCsi(Csi):
    """Implementation of the Cognitive System Interface (CSI) that gets injected to skills at runtime.

    Responsible to tranlate between the types we expose in the SDK and the types in the `wit.imports` module,
    which are automatically generated from the WIT world via `componentize-py`.

    All parameters to any methods on the `WitCsi` must be type checked, as otherwise `componentize-py`
    produces a non-helpful error message. This is done by using `pydantic.dataclasses`. See the
    docstring of `csi` module for more information.
    """

    def completion_stream(
        self, model: str, prompt: str, params: CompletionParams
    ) -> CompletionStreamResponse:
        request = completion_request_to_wit(CompletionRequest(model, prompt, params))
        stream = wit_inference.CompletionStream(request)
        return WitCompletionStreamResponse(stream)

    def chat_stream(
        self, model: str, messages: list[Message], params: ChatParams
    ) -> ChatStreamResponse:
        request = chat_request_to_wit(ChatRequest(model, messages, params))
        stream = wit_inference.ChatStream(request)
        return WitChatStreamResponse(stream)

    def complete_concurrent(
        self, requests: Sequence[CompletionRequest]
    ) -> list[Completion]:
        wit_requests = [completion_request_to_wit(r) for r in requests]
        completions = wit_inference.complete(wit_requests)
        return [completion_from_wit(completion) for completion in completions]

    def chat_concurrent(self, requests: Sequence[ChatRequest]) -> list[ChatResponse]:
        wit_requests = [chat_request_to_wit(r) for r in requests]
        responses = wit_inference.chat(wit_requests)
        return [chat_response_from_wit(response) for response in responses]

    def explain_concurrent(
        self, requests: Sequence[ExplanationRequest]
    ) -> list[list[TextScore]]:
        wit_requests = [explanation_request_to_wit(r) for r in requests]
        responses = wit_inference.explain(wit_requests)
        return [
            [text_score_from_wit(score) for score in scores] for scores in responses
        ]

    def chunk_concurrent(self, requests: Sequence[ChunkRequest]) -> list[list[Chunk]]:
        wit_requests = [chunk_request_to_wit(r) for r in requests]
        responses = wit_chunking.chunk_with_offsets(wit_requests)
        return [[chunk_from_wit(chunk) for chunk in response] for response in responses]

    def select_language_concurrent(
        self, requests: Sequence[SelectLanguageRequest]
    ) -> list[Language | None]:
        wit_requests = [language_request_to_wit(r) for r in requests]
        languages = wit_language.select_language(wit_requests)
        return [
            language_from_wit(lang) if lang is not None else None for lang in languages
        ]

    def search_concurrent(
        self, requests: Sequence[SearchRequest]
    ) -> list[list[SearchResult]]:
        wit_requests = [search_request_to_wit(r) for r in requests]
        results = wit_document_index.search(wit_requests)
        return [
            [search_result_from_wit(result) for result in results_per_request]
            for results_per_request in results
        ]

    def documents(self, document_paths: Sequence[DocumentPath]) -> list[Document]:
        requests = [document_path_to_wit(path) for path in document_paths]
        documents = wit_document_index.documents(requests)
        return [document_from_wit(document) for document in documents]

    def documents_metadata(
        self, document_paths: Sequence[DocumentPath]
    ) -> list[JsonSerializable]:
        requests = [document_path_to_wit(path) for path in document_paths]
        metadata = wit_document_index.document_metadata(requests)
        return [json.loads(metadata) if metadata else None for metadata in metadata]



from typing import Sequence

from pydantic import RootModel

from pharia_skill.csi import Chunk, ChunkRequest

ChunkRequestSerializer = RootModel[Sequence[ChunkRequest]]


ChunkDeserializer = RootModel[list[list[Chunk]]]



import inspect
import json
import traceback
from typing import Callable, Type, TypeVar

from pydantic import BaseModel

from pharia_skill import Csi
from pharia_skill.message_stream.writer import MessageWriter, Payload

UserInput = TypeVar("UserInput", bound=BaseModel)


def message_stream(
    func: Callable[[Csi, MessageWriter[Payload], UserInput], None],
) -> Callable[[Csi, MessageWriter[Payload], UserInput], None]:
    """Turn a function with a specific signature into a (streaming) skill that can be deployed on Pharia Kernel.

    By using the response object, a Skill decorated with `@message_stream` can return intermediate results
    that are streamed to the caller.

    The decorated function must be typed. It must have exactly three arguments. The first argument
    must be of type `Csi`. The second argument must be a `Response` object. The third argument
    must be a Pydantic model. The function must not return anything.

    Example::

        from pharia_skill import Csi, ChatParams, Message, message_stream, MessageWriter
        from pydantic import BaseModel
        from pharia_skill.csi.inference import FinishReason

        class Input(BaseModel):
            topic: str

        class SkillOutput(BaseModel):
            finish_reason: FinishReason

        @message_stream
        def haiku_stream(csi: Csi, writer: MessageWriter[SkillOutput], input: Input) -> None:
            model = "llama-3.1-8b-instruct"
            messages = [
                Message.system("You are a poet who strictly speaks in haikus."),
                Message.user(input.topic),
            ]
            params = ChatParams()
            with csi.chat_stream(model, messages, params) as response:
                writer.begin_message()
                for event in response.stream():
                    writer.append_to_message(event.content)
                writer.end_message(SkillOutput(finish_reason=response.finish_reason()))
    """
    # The import is inside the decorator to ensure the imports only run when the decorator is interpreted.
    # This is because we can only import them when targeting the `message-stream-skill` world.
    # If we target the `skill` world with a component and have the imports for the `message-stream-skill` world
    # in this module at the top-level, we will get a build error in case this module is in the module graph.
    from pharia_skill.bindings import exports
    from pharia_skill.bindings.exports.message_stream import (
        Error_Internal,
        Error_InvalidInput,
    )
    from pharia_skill.bindings.imports import streaming_output as wit
    from pharia_skill.bindings.types import Err
    from pharia_skill.message_stream.wit_writer import WitMessageWriter
    from pharia_skill.wit_csi.csi import WitCsi

    signature = list(inspect.signature(func).parameters.values())
    assert len(signature) == 3, (
        "Message Stream Skills must have exactly three arguments."
    )

    input_model: Type[UserInput] = signature[2].annotation
    assert isinstance(input_model, type) and issubclass(input_model, BaseModel), (
        "The third argument must be a Pydantic model, found: " + str(input_model)
    )

    assert func.__annotations__.get("return") is None, (
        "The function must not return anything"
    )

    # We don't require the schema of the end payload in the function definition.
    # The only use case for this would be to know the metadata of the end payload.
    # Since we don't do metadata for streaming skills at the moment, it is not needed.

    class MessageStream(exports.MessageStream):
        def run(self, input: bytes, output: wit.StreamOutput) -> None:
            """This is the function that gets executed when running the Skill as a Wasm component."""
            try:
                validated = input_model.model_validate_json(input)
            except Exception:
                raise Err(Error_InvalidInput(traceback.format_exc()))
            try:
                with WitMessageWriter[Payload](output) as writer:
                    func(WitCsi(), writer, validated)
            except Exception:
                raise Err(Error_Internal(traceback.format_exc()))

    assert "MessageStream" not in func.__globals__, (
        "Make sure to decorate with `@message_stream` only once."
    )

    def trace_message_stream(
        csi: Csi, writer: MessageWriter[Payload], input: UserInput
    ) -> None:
        """This is the function that we return from the decorator and that get's executed at test time.

        The `opentelemetry` library import is moved to within the function to not make it a dependency of the
        Wasm component.
        """
        from opentelemetry import trace

        with trace.get_tracer(__name__).start_as_current_span(func.__name__) as span:
            writer.span = span  # type: ignore

            span.set_attribute("input", json.dumps(input.model_dump()))
            func(csi, writer, input)
            return

    func.__globals__["MessageStream"] = MessageStream
    trace_message_stream.__globals__["MessageStream"] = MessageStream
    return trace_message_stream



"""
A message represents one turn in a conversation with an LLM.

1. To start a conversation with an LLM, a developer creates a user and optionally system message: `UserMessage(content)`.
2. The LLM responds with an `AssistantMessage` which can include tool calls.
3. If the LLM has requested a tool call, the developer executes the tool call and responds with a `ToolResponse`.
"""

import datetime as dt
import json
from dataclasses import dataclass
from enum import Enum
from typing import Literal, Sequence

from .response import RawResponse, Response, SpecialTokens
from .tool import BuiltInTools, CodeInterpreter, JsonSchema, Tool, ToolDefinition
from .tool_call import ToolCall


class Role(str, Enum):
    """A role used for a message in a chat."""

    User = "user"
    Assistant = "assistant"
    System = "system"
    IPython = "ipython"

    def render(self) -> str:
        return f"{SpecialTokens.StartHeader.value}{self.value.lower()}{SpecialTokens.EndHeader.value}"


@dataclass
class UserMessage:
    """Describes a user message in a chat.

    Parameters:
        content (str, required): The content of the message.
    """

    content: str
    role: Literal[Role.User] = Role.User

    def render(self, tools: Sequence[ToolDefinition]) -> str:
        return f"{Role.User.render()}\n\n{self.content}{SpecialTokens.EndOfTurn.value}"

    @staticmethod
    def json_based_tools(tools: Sequence[ToolDefinition]) -> Sequence[ToolDefinition]:
        """Tools that are defined as JSON schema and invoked with json based tool calling.

        We insert these in the user prompt. The model card states:

        The tool definition is provided in the user prompt, as that is how the model was
        trained for the built in JSON tool calling. However, it's possible to provide
        the tool definition in the system prompt as well—and get similar results.
        Developers must test which way works best for their use case.
        """
        return [tool for tool in tools if tool not in BuiltInTools]


@dataclass
class SystemMessage:
    """Describes a system message in a chat.

    This class is not exposed to the user and is not part of the conversation history like other messages.
    Instead, there is an optional `system` field on the `ChatRequest` which the user can use to set the system message.
    For rendering of the chat history, we then create a `SystemMessage` with the content of the `system` field.

    Parameters:
        content (str, required): The content of the message.
    """

    content: str
    role: Literal[Role.System] = Role.System

    def __init__(self, content: str):
        self.content = content

    def render(self, tools: Sequence[ToolDefinition]) -> str:
        """Render a system message and inject tools into the prompt.

        Always activate the IPython environment if any tools are provided. Activating
        this environment is optional in case there is only user-defined tools. However,
        eval shows that the tool call quality for json based tools is better when the
        IPython environment is activated.

        If built in tools are configured, they are listed in the system prompt.
        The code interpreter tools is automatically included when IPython is activated.

        Reference: https://github.com/meta-llama/llama-models/blob/main/models/llama3_3/prompt_format.md#input-prompt-format-2
        """

        def render_content(content: str) -> str:
            return f"{Role.System.render()}\n\n{content}{SpecialTokens.EndOfTurn.value}"

        def render_tool(tool: ToolDefinition) -> str:
            schema = (
                tool.model_dump()
                if isinstance(tool, JsonSchema)
                else tool.json_schema()
            )
            return json.dumps(schema, indent=4)

        if not tools:
            return render_content(self.content)

        # CodeInterpreter is automatically ijncluded when IPython is activated and does not need to be listed in the system prompt.
        content = "Environment: ipython"
        if filtered := self.system_prompt_tools(tools):
            content += f"\nTools: {', '.join(tool.name() for tool in filtered)}"

        if CodeInterpreter in tools:
            content += "\nIf you decide to run python code, assign the result to a variable called `result`."

        content += f"\nCutting Knowledge Date: December 2023\nToday Date: {dt.datetime.now().strftime('%d %B %Y')}"

        if json_tools := self.json_based_tools(tools):
            content += (
                "\n\nAnswer the user's question by making use of the following functions if needed.\n"
                "Only use functions if they are relevant to the user's question.\n"
                "Here is a list of functions in JSON format:\n"
            )
            for tool in json_tools:
                content += f"{render_tool(tool)}\n"
            content += "\nReturn function calls in JSON format."

        content += "\n\nYou are a helpful assistant."

        # include the original system prompt
        if self.content:
            content += f"\n{self.content}"
        return render_content(content)

    @staticmethod
    def json_based_tools(tools: Sequence[ToolDefinition]) -> Sequence[ToolDefinition]:
        """Tools that are defined as JSON schema and invoked with json based tool calling.

        We insert these in the user prompt. The model card states:

        The tool definition is provided in the user prompt, as that is how the model was
        trained for the built in JSON tool calling. However, it's possible to provide
        the tool definition in the system prompt as well—and get similar results.
        Developers must test which way works best for their use case.
        """
        return [tool for tool in tools if tool not in BuiltInTools]

    @staticmethod
    def system_prompt_tools(tools: Sequence[ToolDefinition]) -> list[type[Tool]]:
        """Subset of specified tools that need to be activated in the system prompt.

        CodeInterpreter is automatically included when IPython is activated and does
        not need to be listed in the system prompt.
        """
        return [
            tool
            for tool in tools
            if isinstance(tool, type)
            and tool in BuiltInTools
            and tool != CodeInterpreter
        ]


@dataclass
class ToolMessage:
    """
    Response for the model after a tool call has been executed.

    Given the LLM has requested a tool call and the developer has executed the tool call,
    the result can be passed back to the model as a `ToolResponse`.
    """

    content: str
    role: Literal[Role.IPython] = Role.IPython
    success: bool = True

    def __init__(self, content: str, success: bool = True):
        self.content = content
        self.success = success

    def render(self, tools: Sequence[ToolDefinition]) -> str:
        return f"{self.role.render()}\n\n{self.output()}{SpecialTokens.EndOfTurn.value}"

    def output(self) -> str:
        """Render the output of the tool call.

        The format defined for llama 3.0 seems to work well: https://github.com/meta-llama/llama-models/blob/main/models/llama3/prompt_templates/tool_response.py#L7

        The model card for 3.1 defines a different format `{"output": "..."}` (https://www.llama.com/docs/model-cards-and-prompt-formats/llama3_1/).
        However, this format does not produce good results.
        """
        prompt = "completed" if self.success else "failed"
        if self.success:
            prompt += f"[stdout]{self.content}[/stdout]"
        else:
            prompt += f"[stderr]{self.content}[/stderr]"
        return prompt


@dataclass
class AssistantMessage:
    """A message that is returned from the LLM."""

    content: str | None = None
    role: Literal[Role.Assistant] = Role.Assistant
    tool_calls: list[ToolCall] | None = None

    def render(self, tools: Sequence[ToolDefinition]) -> str:
        """Always end in <|eom_id|> for tool calls because we always activate the IPython environment.

        Llama will end messages with <|eom_id|> instead of <|eot_id|> if it responds
        with a tool call and `Environment: ipython` is set in the system prompt. If `ipython`
        is not turned on, it will also end tool calls with <|eot_id|>.

        Reference: https://www.llama.com/docs/model-cards-and-prompt-formats/llama3_1/
        """

        def render_content(
            content: str,
            end: Literal[SpecialTokens.EndOfTurn, SpecialTokens.EndOfMessage],
        ) -> str:
            return f"{Role.Assistant.render()}\n\n{content}{end.value}"

        if not self.tool_calls:
            assert self.content is not None, "Content must be set if no tool calls."
            return render_content(self.content, SpecialTokens.EndOfTurn)

        content = SpecialTokens.PythonTag.value
        content += "".join([tool_call.render() for tool_call in self.tool_calls])

        return render_content(content, SpecialTokens.EndOfMessage)

    @staticmethod
    def from_raw_response(
        raw: RawResponse, tools: Sequence[ToolDefinition] | None = None
    ) -> "AssistantMessage":
        response = Response.from_raw(raw)
        if tools and (tool_call := ToolCall.from_response(response, tools)):
            return AssistantMessage(tool_calls=[tool_call])
        return AssistantMessage(content=response.text)



import os
from collections.abc import Sequence
from typing import Optional
from urllib.parse import urljoin

import requests
from dotenv import load_dotenv
from pydantic import BaseModel
from requests.exceptions import ConnectionError, MissingSchema

from pharia_skill.studio.exporter import SpanClient
from pharia_skill.studio.span import StudioSpan, StudioSpanList


class StudioProject(BaseModel):
    name: str
    description: Optional[str]


class StudioClient(SpanClient):
    """Client for communicating with Pharia Studio.

    The Studio instance is determined by the environment variable `PHARIA_STUDIO_ADDRESS`.

    Attributes:
      project_id (int, required): The unique identifier of the project currently in use.
    """

    def __init__(
        self,
        project_name: str,
    ) -> None:
        """Initializes the client.

        Runs a health check to check for a valid url of the Studio connection.
        It does not check for a valid authentication token, which happens later.

        Args:
            project_name (str, required): The human readable identifier provided by the user.
        """
        load_dotenv()
        self._token = os.environ["PHARIA_AI_TOKEN"]
        self._headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {self._token}",
        }

        self.url = os.environ["PHARIA_STUDIO_ADDRESS"]

        self._check_connection()

        self._project_name = project_name
        self._project_id: str | None = None

    @classmethod
    def with_project(cls, project_name: str) -> "StudioClient":
        """Set up a client for a project.

        Will create the project if it does not exist.
        """
        studio_client = StudioClient(project_name=project_name)
        if (project_id := studio_client._get_project(project_name)) is None:
            project_id = studio_client.create_project(project_name)

        assert project_id is not None
        studio_client._project_id = project_id
        return studio_client

    def _check_connection(self) -> None:
        url = urljoin(self.url, "/health")
        try:
            response = requests.get(
                url,
                headers=self._headers,
            )
        except MissingSchema:
            raise ValueError(
                "The given url of the studio client is invalid. Make sure to include http:// in your url."
            ) from None
        except ConnectionError:
            raise ValueError(
                "The given url of the studio client does not point to a server."
            ) from None
        try:
            response.raise_for_status()
        except requests.HTTPError:
            raise ValueError(
                f"The given url of the studio client does not point to a healthy studio: {response.status_code}: {response.text}"
            ) from None

    @property
    def project_id(self) -> str:
        "The unique project_id for the project_name as assigned by Pharia Studio"
        if self._project_id is None:
            if (project_id := self._get_project(self._project_name)) is None:
                raise ValueError(
                    f"Project {self._project_name} was not available. Consider creating it with `StudioClient.create_project`."
                )
            self._project_id = project_id
        return self._project_id

    def _get_project(self, project: str) -> str | None:
        url = urljoin(self.url, "/api/projects")
        response = requests.get(
            url,
            headers=self._headers,
        )
        response.raise_for_status()
        all_projects = response.json()
        try:
            project_of_interest = next(
                proj for proj in all_projects if proj["name"] == project
            )
            return str(project_of_interest["project_id"])
        except StopIteration:
            return None

    def create_project(self, project: str, description: Optional[str] = None) -> str:
        """Creates a project in Studio.

        Projects are uniquely identified by the user provided name.

        Args:
            project (str, required): User provided name of the project.
            description (str, optional, default None): Description explaining the usage of the project.

        Returns:
            The ID of the newly created project.
        """
        url = urljoin(self.url, "/api/projects")
        data = StudioProject(name=project, description=description)
        response = requests.post(
            url,
            data=data.model_dump_json(),
            headers=self._headers,
        )
        match response.status_code:
            case 409:
                raise ValueError("Project already exists")
            case _:
                response.raise_for_status()
        return response.text

    def submit_spans(self, spans: Sequence[StudioSpan]) -> None:
        """Sends the provided spans to Studio as a singular trace.

        The method fails if the span list is empty, has already been created or if
        spans belong to multiple traces.

        Args:
            spans (Sequence[StudioSpan], required): Spans to create the trace from. Created by exporting from a :class:`Tracer`.
        """
        if len(spans) == 0:
            raise ValueError("Tried to upload an empty trace")
        self._upload_trace(StudioSpanList(spans))

    def _upload_trace(self, trace: StudioSpanList) -> None:
        url = urljoin(self.url, f"/api/projects/{self.project_id}/traces")
        response = requests.post(
            url,
            data=trace.model_dump_json(),
            headers=self._headers,
        )
        match response.status_code:
            case 409:
                raise ValueError(
                    f"Trace with id {trace.root[0].context.trace_id} already exists."
                )
            case 422:
                raise ValueError(
                    f"Uploading the trace failed with 422. Response: {response.json()}"
                )
            case _:
                response.raise_for_status()



import io
import logging
import os
import platform
import subprocess
import sys
import tarfile
import zipfile
from pathlib import Path
from typing import NamedTuple

import requests
from dotenv import load_dotenv

logger = logging.getLogger(__name__)


class Registry(NamedTuple):
    """Where and How do I publish my skill?"""

    user: str
    token: str
    registry: str
    repository: str

    @classmethod
    def from_env(cls) -> "Registry":
        return cls(
            user=os.environ["SKILL_REGISTRY_USER"],
            token=os.environ["SKILL_REGISTRY_TOKEN"],
            registry=os.environ["SKILL_REGISTRY"],
            repository=os.environ["SKILL_REPOSITORY"],
        )


class PhariaSkillCli:
    """The `pharia-skill-cli` rust crate is used for publishing skills.

    This class manages the installation of the `pharia-skill-cli` binary and provides
    an interface to its commands.

    We make sure the `pharia-skill-cli` binary is up to date before allowing users to invoke commands.
    """

    # Expected version of the `pharia-skill-cli` binary
    PHARIA_SKILL_CLI_VERSION = "0.4.4"

    PHARIA_SKILL_CLI_PATH = (
        "bin/pharia-skill-cli"
        if "Windows" not in platform.system()
        else ".\\bin\\pharia-skill-cli.exe"
    )

    def __init__(self) -> None:
        load_dotenv()
        self.update_if_needed()

    @classmethod
    def update_if_needed(cls) -> None:
        if not os.path.exists(cls.PHARIA_SKILL_CLI_PATH) or not cls.is_up_to_date():
            cls.download_pharia_skill()
            assert cls.is_up_to_date()

    @classmethod
    def is_up_to_date(cls) -> bool:
        return cls.pharia_skill_version() == cls.PHARIA_SKILL_CLI_VERSION

    @classmethod
    def pharia_skill_version(cls) -> str | None:
        """Version of the currently installed `pharia-skill-cli` binary."""
        result = subprocess.run(
            [cls.PHARIA_SKILL_CLI_PATH, "--version"],
            stdout=subprocess.PIPE,
            text=True,
        )
        if result.returncode != 0:
            return None

        # pharia-skill-cli version is in the format "pharia-skill-cli 0.1.0"
        return result.stdout.strip().split(" ")[-1]

    @classmethod
    def architecture(cls) -> str:
        match platform.system():
            case "Darwin":
                if platform.machine() == "arm64":
                    return "aarch64-apple-darwin"
                else:
                    return "x86_64-apple-darwin"
            case "Linux":
                return "x86_64-unknown-linux-gnu"
            case "Windows":
                return "x86_64-pc-windows-msvc"
            case _:
                raise Exception(f"Unsupported operating system: {platform.system()}")

    @classmethod
    def download_pharia_skill(cls) -> None:
        logger.info(
            f"Downloading pharia-skill-cli version {cls.PHARIA_SKILL_CLI_VERSION} for {cls.architecture()}"
        )
        os.makedirs("bin", exist_ok=True)
        for file in os.listdir("bin"):
            os.remove(os.path.join("bin", file))

        match cls.architecture():
            case (
                "aarch64-apple-darwin"
                | "x86_64-apple-darwin"
                | "x86_64-unknown-linux-gnu"
            ):
                cls.download_unix_tar(Path("bin"))
            case "x86_64-pc-windows-msvc":
                cls.download_windows_zip(Path("bin"))
            case _:
                raise Exception(f"Unsupported architecture: {cls.architecture()}")

        logger.info("Pharia skill CLI installed successfully.")

    @classmethod
    def download_unix_tar(cls, dir: Path) -> None:
        filename = f"pharia-skill-cli-{cls.architecture()}"
        url = f"https://github.com/Aleph-Alpha/pharia-skill-cli/releases/download/v{cls.PHARIA_SKILL_CLI_VERSION}/{filename}.tar.xz"
        response = requests.get(url)
        if response.status_code != 200:
            raise Exception(f"{response.status_code}: {response.text}")

        file = tarfile.open(fileobj=io.BytesIO(response.content)).extractfile(
            filename + "/pharia-skill-cli"
        )
        assert file, "pharia-skill-cli not found in archive"
        with open(dir / "pharia-skill-cli", "wb") as f:
            f.write(file.read())
        subprocess.run(["chmod", "+x", dir / "pharia-skill-cli"], check=True)

    @classmethod
    def download_windows_zip(cls, dir: Path) -> None:
        url = f"https://github.com/Aleph-Alpha/pharia-skill-cli/releases/download/v{cls.PHARIA_SKILL_CLI_VERSION}/pharia-skill-cli-x86_64-pc-windows-msvc.zip"
        print(url)
        response = requests.get(url)
        if response.status_code != 200:
            raise Exception(f"{response.status_code}: {response.text}")

        zipfile.ZipFile(io.BytesIO(response.content)).extract(
            "pharia-skill-cli.exe", path=dir
        )

    def publish(
        self, skill: str, name: str | None, tag: str, registry: Registry
    ) -> None:
        """Publish a skill to an OCI registry.

        Takes a path to a Wasm component, wrap it in an OCI image and publish it to an OCI
        registry under the `latest` tag. This does not fully deploy the skill, as an older
        version might still be cached in the Kernel.
        """
        # add file extension
        if not skill.endswith(".wasm"):
            skill += ".wasm"

        # allow relative paths
        if not skill.startswith(("/", "./")):
            skill = f"./{skill}"

        if not os.path.exists(skill):
            logger.error(f"No such file: {skill}")
            sys.exit(1)

        command = [
            self.PHARIA_SKILL_CLI_PATH,
            "publish",
            "-R",
            registry.registry,
            "-r",
            registry.repository,
            "-u",
            registry.user,
            "-p",
            registry.token,
            *(["-n", name] if name else []),
            "-t",
            tag,
            skill,
        ]

        try:
            subprocess.run(command, check=True)
        except subprocess.CalledProcessError:
            sys.exit(1)



import inspect
import json
import traceback
from typing import Callable, Type, TypeVar

# from opentelemetry import trace
# from opentelemetry.trace import Status, StatusCode
from pydantic import (
    BaseModel,
    # For generation of JSON schemas, Pydantic imports the `root_model` module at runtime: https://github.com/pydantic/pydantic/blob/main/pydantic/json_schema.py#L1500
    # As `componentize-py` resolves imports at build time, we are required to add this import here.
    RootModel,  # noqa: F401
)

from .bindings import exports
from .bindings.types import Err
from .csi import Csi
from .wit_csi import WitCsi

UserInput = TypeVar("UserInput", bound=BaseModel)
UserOutput = TypeVar("UserOutput", bound=BaseModel)


def skill(
    func: Callable[[Csi, UserInput], UserOutput],
) -> Callable[[Csi, UserInput], UserOutput]:
    """Turn a function with a specific signature into a skill that can be deployed on Pharia Kernel.

    The decorated function must be typed. It must have exactly two input arguments. The first argument
    must be of type `Csi`. The second argument must be a Pydantic model. The type of the return value
    must also be a Pydantic model. Each module is expected to have only one function that is decorated
    with `skill`.

    Example::

        from pharia_skill import ChatParams, Csi, Message, skill
        from pydantic import BaseModel

        class Input(BaseModel):
            topic: str

        class Output(BaseModel):
            haiku: str

        @skill
        def run(csi: Csi, input: Input) -> Output:
            system = Message.system("You are a poet who strictly speaks in haikus.")
            user = Message.user(input.topic)
            params = ChatParams(max_tokens=64)
            response = csi.chat("llama-3.1-8b-instruct", [system, user], params)
            return Output(haiku=response.message.content.strip())
    """
    # The import is inside the decorator to ensure the imports only run when the decorator is interpreted.
    # This is because we can only import them when targeting the `skill` world.
    # If we target the `message-stream-skill` world with a component and have the imports for the `skill` world
    # in this module at the top-level, we will get a build error in case this module is in the module graph.
    from .bindings.exports.skill_handler import (
        Error_Internal,
        Error_InvalidInput,
        SkillMetadata,
    )

    signature = list(inspect.signature(func).parameters.values())
    assert len(signature) == 2, "Skills must have exactly two arguments."

    input_model: Type[UserInput] = signature[1].annotation
    assert issubclass(input_model, BaseModel), (
        "The second argument must be a Pydantic model"
    )

    assert func.__annotations__.get("return") is not None, (
        "The function must have a return type annotation"
    )
    output_model: Type[UserOutput] = func.__annotations__["return"]
    assert issubclass(output_model, BaseModel), (
        "The return type must be a Pydantic model"
    )

    # This code here inside the decorator (but outside of the `class SkillHandler`) is executed at build time.
    # In version 0.3 of the wit world, we did not account for the fact that the metadata method may return
    # an error. However, as pydantic does some imports at runtime, we need to take this possibility into account.
    # By calculating the metadata at build time, we can (in case there is an error) give the user direct feedback,
    # instead of failing at runtime.
    description = func.__doc__
    input_schema = json.dumps(input_model.model_json_schema()).encode()
    output_schema = json.dumps(output_model.model_json_schema()).encode()
    metadata = SkillMetadata(description, input_schema, output_schema)

    class SkillHandler(exports.SkillHandler):
        def run(self, input: bytes) -> bytes:
            """This is the function that gets executed when running the Skill as a Wasm component."""
            try:
                validated = input_model.model_validate_json(input)
            except Exception:
                raise Err(Error_InvalidInput(traceback.format_exc()))
            try:
                result = func(WitCsi(), validated)
                return result.model_dump_json().encode()
            except Exception:
                raise Err(Error_Internal(traceback.format_exc()))

        def metadata(self) -> SkillMetadata:
            return metadata

    assert "SkillHandler" not in func.__globals__, "`@skill` can only be used once."

    def trace_skill(csi: Csi, input: UserInput) -> UserOutput:
        """This is the function that we return from the decorator and that get's executed at test time.

        The `opentelemetry` library import is moved to within the function to not make it a dependency of the
        Wasm component.
        """
        from opentelemetry import trace

        with trace.get_tracer(__name__).start_as_current_span(func.__name__) as span:
            span.set_attribute("input", json.dumps(input.model_dump()))
            result = func(csi, input)
            span.set_attribute("output", json.dumps(result.model_dump()))
            return result

    func.__globals__["SkillHandler"] = SkillHandler
    trace_skill.__globals__["SkillHandler"] = SkillHandler
    return trace_skill



"""
Make HTTP requests against a running Pharia Kernel.

By separating the client from the `DevCsi`, we can better test the serialization/deserialization
and other functionality of the `DevCsi` without making HTTP requests.
"""

import json
import os
from http import HTTPStatus
from typing import Any, Generator, Protocol

import requests
from dotenv import load_dotenv
from pydantic import BaseModel


class Event(BaseModel):
    event: str
    data: dict[str, Any]


class CsiClient(Protocol):
    def run(self, function: str, data: dict[str, Any]) -> Any: ...
    def stream(
        self, function: str, data: dict[str, Any]
    ) -> Generator[Event, None, None]: ...


class Client(CsiClient):
    """Make requests with a given payload against a running Pharia Kernel."""

    HTTP_CSI_VERSION = "v1"

    def __init__(self) -> None:
        """Create a new HTTP client.

        The session is stored to allow for re-use of the same connection between tests.
        """
        load_dotenv()
        self.url = (
            f"""{os.environ["PHARIA_KERNEL_ADDRESS"]}/csi/{self.HTTP_CSI_VERSION}"""
        )
        token = os.environ["PHARIA_AI_TOKEN"]
        self.session = requests.Session()
        self.session.headers = {
            "Authorization": f"Bearer {token}",
        }

    def __del__(self) -> None:
        if hasattr(self, "session"):
            self.session.close()

    def run(self, function: str, data: Any) -> Any:
        url = f"{self.url}/{function}"
        response = self.session.post(
            url,
            json=data,
        )
        # Always forward the error message from the kernel
        if response.status_code >= 400:
            try:
                error = response.json()
            except requests.JSONDecodeError:
                error = response.text
            raise Exception(
                f"{response.status_code} {HTTPStatus(response.status_code).phrase}: {error}"
            )

        return response.json()

    def stream(
        self, function: str, data: dict[str, Any]
    ) -> Generator[Event, None, None]:
        url = f"{self.url}/{function}"
        headers = {"Accept": "text/event-stream", **self.session.headers}
        response = self.session.post(url, json=data, headers=headers, stream=True)
        # Always forward the error message from the kernel
        if response.status_code >= 400:
            try:
                error = response.json()
            except requests.JSONDecodeError:
                error = response.text
            raise Exception(
                f"{response.status_code} {HTTPStatus(response.status_code).phrase}: {error}"
            )

        return KernelStreamDeserializer(response).events()


class KernelStreamDeserializer:
    def __init__(self, response: requests.Response) -> None:
        self.response = response

    def _read(self) -> Generator[bytes, None, None]:
        """Read the incoming event source stream and yield event chunks.

        Unfortunately it is possible for some servers to decide to break an
        event into multiple HTTP chunks in the response. It is thus necessary
        to correctly stitch together consecutive response chunks and find the
        SSE delimiter (empty new line) to yield full, correct event chunks."""
        data = b""
        for chunk in self.response:
            for line in chunk.splitlines(True):
                data += line
                if data.endswith((b"\r\r", b"\n\n", b"\r\n\r\n")):
                    yield data
                    data = b""
        if data:
            yield data

    def events(self) -> Generator[Event, None, None]:
        """Yield events from the kernel stream.

        This method does not raise on error events, but rather deserializes them and
        leaves it to the caller to raise an exception.
        """
        for stream_item in self._read():
            text = stream_item.decode().strip()
            if not text:
                continue

            # Kernel events always have this format:
            # event: <event>
            # data: <data>
            if "\n" not in text:
                raise ValueError(f"Unexpected event format: {text}")
            first_line, remaining = text.split("\n", 1)

            if not first_line.startswith("event: "):
                raise ValueError(f"Unexpected event prefix: {first_line}")
            event = first_line.split("event: ", 1)[1]

            if not remaining.startswith("data: "):
                raise ValueError(f"Unexpected data prefix: {remaining}")
            data = remaining.split("data: ", 1)[1]

            yield Event(event=event, data=json.loads(data))
